#include <proto.h>

#include "gtgroup.h"
#include "gtapp.h"
#include "gterror.h"
#include "gtresname.h"
#include "gtcommon.h"

/************************************************************************\
*
* GtGroup class.
*
* $Ident: This file is part of the 'Gt-library' 1.0.0 $
* $Copyright: (C) 1997 by Gran Gustafsson $
*
\************************************************************************/

/************************************************************************\
*
* Create and init group.
*
\************************************************************************/
GtGroup *GtMkGroup(char *n,GtManager *p,GtTag *tgs)
{
	GtGroup *obj;
	
	if(!(obj=new GtGroup()))
	{
		GtError(NULL,GTEC_NOOBJECT,"No GtGroup object!\n");
		return(NULL);
	}
	if(!obj->Init(n,p,tgs))
	{
		GtDestroy(obj);
		obj=NULL;
	}
	return(obj);
}
	
/************************************************************************\
*
* Create and init group.
*
\************************************************************************/
GtGroup *GtMkGroup(char *n,GtManager *p,ulong tag,...)
{
	va_list va;
	GtGroup *obj;
	int ps;
	
	va_start(va,tag);
	ps=GtBeginVaTags();
	obj=GtMkGroup(n,p,GtCopyVaTags(tag,&va));
	GtEndVaTags(ps);
	va_end(va);
	return(obj);
}
	
/************************************************************************\
*
* Setup empty group.
*
\************************************************************************/
GtGroup::GtGroup(void)
{
	totw=0;
	expsize=1;
	dir=GTO_HORIZ;
	ftype=GTF_NONE;
	ibw=0;
	ibdrw=0;
	bdrw=1;
	space=0;
}

/************************************************************************\
*
* Setup group.
*
\************************************************************************/
sbool GtGroup::Init(char *n,GtManager *p,GtTag *tgs)
{
	GtTag *t;
	
	if(!GtWhtMan::Init(n,p,tgs))
		return(FALSE);
	app->GetRes(this,GTR_ORIENT,GTRT_ORIENT,&dir);
	app->GetRes(this,GTR_FRAME,GTRT_FRAME,&ftype);
	app->GetRes(this,GTR_SPACE,GTRT_BYTE,&space);
	app->GetRes(this,GTR_IBDR,GTRT_BYTE,&ibdrw);
	app->GetRes(this,GTR_BDR,GTRT_BYTE,&bdrw);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Orient:
				dir=t->d.val;
				if(dir>GTO_LAST_NR)
					dir=GTO_VERT;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				if(ftype>GTF_LAST_NR)
					ftype=GTF_NONE;
				break;
			case GT_Spacing:
				space=t->d.val;
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				break;
			case GT_Border:
				bdrw=t->d.val;
				break;
		}
	}
	switch(ftype)
	{
		case GTF_NONE:
			ibw=ibdrw;
			break;
		case GTF_SIMPLE:
		case GTF_IN:
		case GTF_OUT:
			ibw=ibdrw+bdrw;
			break;
		case GTF_DBL_IN:
		case GTF_DBL_OUT:
			ibw=ibdrw+2*bdrw;
			break;
	}
	this->SetMinMaxSize(GTMAX(2*ibw+1,1),GTMAX(2*ibw+1,1),GTS_MAX,GTS_MAX);
	return(TRUE);
}

/************************************************************************\
*
* Set group attributes.
*
\************************************************************************/
sbool GtGroup::Set(GtTag *tgs)
{
	GtTag *t;
	sbool ch=FALSE;
	
	if(!GtWhtMan::Set(tgs))
		return(FALSE);
	while((t=GtNextTag(&tgs)))
	{
		switch(t->tag)
		{
			case GT_Orient:
				dir=t->d.val;
				if(dir>GTO_LAST_NR)
					dir=GTO_VERT;
				ch=TRUE;
				break;
			case GT_FrameType:
				ftype=t->d.val;
				if(ftype>GTF_LAST_NR)
					ftype=GTF_NONE;
				ch=TRUE;
				break;
			case GT_Spacing:
				space=t->d.val;
				ch=TRUE;
				break;
			case GT_InnerBorder:
				ibdrw=t->d.val;
				ch=TRUE;
				break;
			case GT_Border:
				bdrw=t->d.val;
				ch=TRUE;
				break;
		}
	}
	if(ch)
	{
		switch(ftype)
		{
			case GTF_NONE:
				ibw=ibdrw;
				break;
			case GTF_SIMPLE:
			case GTF_IN:
			case GTF_OUT:
				ibw=ibdrw+bdrw;
				break;
			case GTF_DBL_IN:
			case GTF_DBL_OUT:
				ibw=ibdrw+2*bdrw;
				break;
		}
		this->Relayout();
	}
	return(TRUE);
}
	
/************************************************************************\
*
* Get group attributes.
*
\************************************************************************/
void GtGroup::Get(GtTag *tgs)
{
	GtTag *t;
	long *ptr;
	
	GtWhtMan::Get(tgs);
	while((t=GtNextTag(&tgs)))
	{
		if(!(ptr=(long *)t->d.ptr))
			continue;
		switch(t->tag)
		{
			case GT_Orient:
				*ptr=dir;
				break;
			case GT_FrameType:
				*ptr=ftype;
				break;
			case GT_Spacing:
				*ptr=space;
				break;
			case GT_InnerBorder:
				*ptr=ibdrw;
				break;
			case GT_Border:
				*ptr=bdrw;
				break;
		}
	}
}
	
/************************************************************************\
*
* Check if class is (sub)class of named class.
*
\************************************************************************/
sbool GtGroup::IsSubClassOf(char *cn)
{
	if(strcmp(cn,"GtGroup")==0)
		return(TRUE);
	else
		return(GtWhtMan::IsSubClassOf(cn));
}

/************************************************************************\
*
* Set group event mask.
*
\************************************************************************/
void GtGroup::SetEMask(ulong em)
{
	GtWhtMan::SetEMask(em|ExposureMask);
}

/************************************************************************\
*
* Relayout group.
*
\************************************************************************/
void GtGroup::Relayout(void)
{
	GtWidget *obj,*tmp,**last;
	GtWListWht *li,*tli;
	int xs,ys,n,mxxs,mxys,xsm,ysm;
	word mxs,mys;
	
	if(flags & GTWF_UNBUILT)
		return;
	mxs=2*ibw;
	mys=2*ibw;
	if(dir==GTO_HORIZ)
	{
		mxxs=0;
		mxys=GTS_MAX;
	}
	else
	{
		mxxs=GTS_MAX;
		mxys=0;
	}
	first=NULL;
	totw=0;
	expsize=0;
	n=0;
	for(li=(GtWListWht *)GtSFirst(&man); li;
			li=(GtWListWht *)GtSNext(&li->bli.wl.lnk))
	{
		obj=li->bli.wl.w;
		if(!obj->IsMapped())
			continue;
		totw+=li->weight;
		obj->GetMinSize(&xs,&ys);
		li->bli.minxs=GTMAX(xs,li->bli.minxs);
		li->bli.minys=GTMAX(ys,li->bli.minys);
		obj->GetMaxSize(&xsm,&ysm);
		li->bli.maxxs=GTMAX(li->bli.minxs,GTMIN(xsm,li->bli.maxxs));
		li->bli.maxys=GTMAX(li->bli.minys,GTMIN(ysm,li->bli.maxys));
		if(dir==GTO_HORIZ)
		{
			mxs+=(li->bli.minxs+li->bli.bleft+li->bli.bright);
			mxxs+=(li->bli.maxxs+li->bli.bleft+li->bli.bright);
			if(GtSNext(&li->bli.wl.lnk))
			{
				mxs+=space;
				mxxs+=space;
			}
			mxxs=GTMIN(mxxs,GTS_MAX);
			mys=GTMAX(mys,li->bli.minys+li->bli.btop+li->bli.bbot+2*ibw);
			mxys=GTMIN(mxys,li->bli.maxys+li->bli.btop+li->bli.btop+2*ibw);
			expsize=GTMAX(expsize,li->bli.minys+li->bli.btop+li->bli.bbot);
			tmp=first;
			last=&first;
			while(tmp)
			{
				tli=(GtWListWht *)tmp->GetLInfo();
				if(li->bli.maxxs-li->bli.minxs<tli->bli.maxxs-tli->bli.minxs)
					break;
				last=&tli->nxt;
				tmp=tli->nxt;
			}
			li->nxt=*last;
			*last=obj;
		}
		else
		{
			mxs=GTMAX(mxs,li->bli.minxs+li->bli.bleft+li->bli.bright+2*ibw);
			mxxs=GTMIN(mxxs,li->bli.maxxs+li->bli.bleft+li->bli.bright+2*ibw);
			mys+=(li->bli.minys+li->bli.btop+li->bli.bbot);
			mxys+=(li->bli.maxys+li->bli.btop+li->bli.bbot);
			if(GtSNext(&li->bli.wl.lnk))
			{
				mys+=space;
				mxys+=space;
			}
			mxys=GTMIN(mxys,GTS_MAX);
			expsize=GTMAX(expsize,li->bli.minxs+li->bli.bleft+li->bli.bright);
			tmp=first;
			last=&first;
			while(tmp)
			{
				tli=(GtWListWht *)tmp->GetLInfo();
				if(li->bli.maxys-li->bli.minys<tli->bli.maxys-tli->bli.minys)
					break;
				last=&tli->nxt;
				tmp=tli->nxt;
			}			
			li->nxt=*last;
			*last=obj;
		}
		n++;
	}
	mxxs=GTMAX(mxs,mxxs);
	mxys=GTMAX(mys,mxys);
	this->SetMinMaxSize(GTMAX(mxs,1+2*ibw),GTMAX(mys,1+2*ibw),mxxs,mxys);
}

/************************************************************************\
*
* Layout group.
*
\************************************************************************/
void GtGroup::Layout(void)
{
	GtWidget *obj;
	GtWListWht *li;
	int xp,yp,xs,ys,nxs,nys,tsz,ctw;
	
	if(flags & GTWF_UNBUILT)
		return;
	if(xsize>minxs)
		xs=xsize-minxs;
	else
		xs=0;
	if(ysize>minys)
		ys=ysize-minys;
	else
		ys=0;
	ctw=totw;
	obj=first;
	while(obj)
	{
		li=(GtWListWht *)obj->GetLInfo();
		if(dir==GTO_HORIZ)
		{
			li->size=li->bli.minxs;
			if(li->weight)
			{
				tsz=xs*li->weight/ctw;
				if(li->bli.maxxs)
					tsz=GTMIN(tsz,li->bli.maxxs-li->size);
				li->size+=tsz;
				xs-=tsz;
				ctw-=li->weight;
			}
		}
		else
		{
			li->size=li->bli.minys;
			if(li->weight)
			{
				tsz=ys*li->weight/ctw;
				if(li->bli.maxys)
					tsz=GTMIN(tsz,li->bli.maxys-li->size);
				li->size+=tsz;
				ys-=tsz;
				ctw-=li->weight;
			}
		}
		obj=li->nxt;
	}
	if(flags & GTWF_NOWIN)
	{
		xp=ibw+xpos;
		yp=ibw+ypos;
	}
	else
	{
		xp=ibw;
		yp=ibw;
	}
	for(li=(GtWListWht *)GtSFirst(&man); li;
			li=(GtWListWht *)GtSNext(&li->bli.wl.lnk))
	{
		obj=li->bli.wl.w;
		if(!obj->IsMapped())
			continue;
		if(dir==GTO_HORIZ)
		{
			xp+=li->bli.bleft;
			nxs=li->size;
			nys=li->bli.minys;
			switch(li->place)
			{
				case GTP_TOPLEFT:
					yp=ibw+li->bli.btop;
					break;
				case GTP_BOTRIGHT:
					yp=ibw+GTMAX(ysize-ibw-nys-li->bli.bbot,li->bli.btop);
					break;
				case GTP_EXPAND:
					tsz=GTMAX(ysize-2*ibw-li->bli.btop-li->bli.bbot,expsize-li->bli.btop-li->bli.bbot);
					if(li->bli.maxys)
						tsz=GTMIN(tsz,li->bli.maxys);
					nys=tsz;
				/* Watch it here !! Dropping through to GTP_CENTER */
				case GTP_CENTER:
					yp=ibw+li->bli.btop+(ysize-2*ibw-nys-li->bli.btop-li->bli.bbot)/2;
					break;
			}
			if(flags & GTWF_NOWIN)
				yp+=ypos;
			obj->SetPosSize(xp,yp,nxs,nys);
			xp+=nxs+li->bli.bright+space;
		}
		else
		{
			yp+=li->bli.btop;
			nys=li->size;
			nxs=li->bli.minxs;
			switch(li->place)
			{
				case GTP_TOPLEFT:
					xp=ibw+li->bli.bleft;
					break;
				case GTP_BOTRIGHT:
					xp=ibw+GTMAX(xsize-ibw-nxs-li->bli.bright,li->bli.bleft);
					break;
				case GTP_EXPAND:
					tsz=GTMAX(xsize-2*ibw-li->bli.bleft-li->bli.bright,expsize-li->bli.bleft-li->bli.bright);
					if(li->bli.maxxs)
						tsz=GTMIN(tsz,li->bli.maxxs);
					nxs=tsz;
				/* Watch it here !! Dropping through to GTP_CENTER */
				case GTP_CENTER:
					xp=ibw+li->bli.bleft+(xsize-2*ibw-nxs-li->bli.bleft-li->bli.bright)/2;
					break;
			}
			if(flags & GTWF_NOWIN)
				xp+=xpos;
			obj->SetPosSize(xp,yp,nxs,nys);
			yp+=nys+li->bli.bbot+space;
		}
	}
}

/************************************************************************\
*
* Draw group.
*
\************************************************************************/
sbool GtGroup::Draw(XEvent *xe)
{
	int n,xp,yp;
	
	if(xe && ((xe->type!=Expose) || (xe->xexpose.count!=0)))
		return(TRUE);
	if(flags & GTWF_NOWIN)
	{
		xp=xpos;
		yp=ypos;
	}
	else
	{
		xp=0;
		yp=0;
	}
	switch(ftype)
	{
		case GTF_SIMPLE:
			for(n=0; n<bdrw; n++)
				gtfore->DrawRectangle(gtwin,xp+n,yp+n,xsize-2*n-1,ysize-2*n-1,FALSE);
			break;
		case GTF_DBL_IN:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtdrk,gtlht,gtwin,xp+n,yp+n,xsize-2*n,ysize-2*n);
				GT3DBOX(gtlht,gtdrk,gtwin,xp+bdrw+n,yp+bdrw+n,
																	xsize-2*(bdrw+n),ysize-2*(bdrw+n));
			}
			break;
		case GTF_DBL_OUT:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtlht,gtdrk,gtwin,xp+n,yp+n,xsize-2*n,ysize-2*n);
				GT3DBOX(gtdrk,gtlht,gtwin,xp+bdrw+n,yp+bdrw+n,
																	xsize-2*(bdrw+n),ysize-2*(bdrw+n));
			}
			break;
		case GTF_IN:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtdrk,gtlht,gtwin,xp+n,yp+n,xsize-2*n,ysize-2*n);
			}
			break;
		case GTF_OUT:
			for(n=0; n<bdrw; n++)
			{
				GT3DBOX(gtlht,gtdrk,gtwin,xp+n,yp+n,xsize-2*n,ysize-2*n);
			}
			break;
	}
	return(TRUE);
}
